dnl  AMD K6 mpn_mullowhalf_basecase -- low half multiply.

dnl  Copyright 2002, 2007 Kevin Ryde
dnl  Copyright 1999, 2000, 2001, 2002 Free Software Foundation, Inc.

dnl  This file is part of MPM.
dnl
dnl  MPM is free software; you can redistribute it and/or modify it under the
dnl  terms of the GNU General Public License as published by the Free Software
dnl  Foundation; either version 2, or (at your option) any later version.
dnl
dnl  MPM is distributed in the hope that it will be useful, but WITHOUT ANY
dnl  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
dnl  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
dnl  details.
dnl
dnl  You should have received a copy of the GNU General Public License along
dnl  with MPM; see the file COPYING.  If not, write to the Free Software
dnl  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
dnl  USA.

include(`../config.m4')
define_mpn(mullowhalf_basecase)

define(MULLOWHALF_THRESHOLD, 66)


C K6: ? cycles/crossproduct




dnl  MULLOWHALF_THRESHOLD_MAX is the maximum MULLOWHALF_THRESHOLD this code
dnl  supports.  This value is used only by the tune program to know what it
dnl  can go up to.  (An attempt to compile with a bigger value will trigger
dnl  some m4_assert()s in the code, making the build fail.)
dnl
dnl  The value is determined by requiring the displacements in the unrolled
dnl  addmul to fit in single bytes.  This means a maximum UNROLL_COUNT of
dnl  63, giving a maximum MULLOWHALF_THRESHOLD of 66.
dnl
dnl  The tuned threshold in fact comes out quite close to 66, but believe
dnl  that limit is not losing any opportunities, at the moment.

deflit(MULLOWHALF_THRESHOLD_MAX, 66)


dnl  Allow a value from the tune program to override config.m4.

ifdef(`MULLOWHALF_THRESHOLD_OVERRIDE',
`define(`MULLOWHALF_THRESHOLD',MULLOWHALF_THRESHOLD_OVERRIDE)')


dnl  UNROLL_COUNT is the number of code chunks in the unrolled addmul.  The
dnl  number required is determined by MULLOWHALF_THRESHOLD, since
dnl  mpn_mullowhalf_basecase only needs to handle sizes < MULLOWHALF_THRESHOLD.
dnl
dnl  The first addmul is the biggest, and this takes the second least
dnl  significant limb and multiplies it by the third least significant and
dnl  up.  Hence for a maximum operand size of MULLOWHALF_THRESHOLD-1 limbs,
dnl  UNROLL_COUNT needs to be MULLOWHALF_THRESHOLD-3.

m4_config_gmp_mparam(`MULLOWHALF_THRESHOLD')
deflit(UNROLL_COUNT, eval(MULLOWHALF_THRESHOLD-3))


C void mpn_mullowhalf_basecase (mp_ptr wp, mp_srcptr xp, mp_srcptr yp,
C                               mp_size_t size);
C

defframe(PARAM_SIZE,16)
defframe(PARAM_YP,  12)
defframe(PARAM_XP,  8)
defframe(PARAM_WP,  4)

	TEXT
	ALIGN(32)
PROLOGUE(mpn_mullowhalf_basecase)
deflit(`FRAME',0)

	movl	PARAM_SIZE, %ecx

	cmpl	$2, %ecx
	movl	PARAM_XP, %eax

	ja	L(three_or_more)
	je	L(two_limbs)

C -----------------------------------------------------------------------------
C size==1

	movl	PARAM_YP, %edx

	movl	(%eax), %eax		C xp[0]

	movl	(%edx), %edx		C yp[0]

	movl	PARAM_WP, %ecx

	imull	%edx

	movl	%eax, (%ecx)

	ret


C -----------------------------------------------------------------------------
	ALIGN(16)
L(two_limbs):
	C eax	xp
	C ebx
	C ecx
	C edx

	movl	PARAM_YP, %ecx
	pushl	%ebx		FRAME_pushl()

	movl	%eax, %ebx		C xp
	movl	(%eax), %edx		C xp[0]

	movl	(%ecx), %eax		C yp[0]
	pushl	%esi		FRAME_pushl()

	mull	%edx			C xp[0] * yp[0]

	movl	4(%ecx), %esi		C yp[1]
	movl	(%ecx), %ecx		C yp[0]

	imull	(%ebx), %esi		C xp[0] * yp[1]

	imull	4(%ebx), %ecx		C xp[1] * yp[0]

	addl	%ecx, %edx
	movl	PARAM_WP, %ecx

	addl	%esi, %edx
	popl	%esi

	movl	%eax, (%ecx)
	popl	%ebx

	movl	%edx, 4(%ecx)

	ret


C -----------------------------------------------------------------------------
L(three_or_more):
deflit(`FRAME',0)
	cmpl	$4, %ecx
	jae	L(four_or_more)


C -----------------------------------------------------------------------------
C three limbs
	C eax	xp
	C ebx
	C ecx
	C edx

	movl	PARAM_YP, %ecx
	pushl	%ebx	FRAME_pushl()

	movl	%eax, %ebx		C xp
	movl	(%eax), %eax		C xp[0]

	movl	(%ecx), %edx		C yp[0]
	pushl	%edi	FRAME_pushl()

	movl	PARAM_WP, %edi
	nop				C code alignment

	pushl	%ebp	FRAME_pushl()
	movl	%eax, %ebp		C xp[0]

	mull	%edx			C xp[0] * yp[0]

	movl	%eax, (%edi)		C wp[0]
	movl	4(%ecx), %eax		C yp[1]

	pushl	%esi	FRAME_pushl()
	movl	%edx, %esi		C wp[1] part

	mull	%ebp			C xp[0] * yp[1]

	imull	8(%ecx), %ebp		C xp[0] * yp[2]

	addl	%eax, %esi		C wp[1] part
	movl	4(%ebx), %eax		C xp[1]

	adcl	%edx, %ebp		C wp[2] part

	movl	(%ecx), %edx		C yp[0]

	mull	%edx			C xp[1] * yp[0]

	addl	%eax, %esi		C wp[1]
	movl	4(%ebx), %eax		C xp[1]

	adcl	%edx, %ebp		C wp[2] part

	movl	8(%ebx), %edx		C xp[0]
	movl	%esi, 4(%edi)		C wp[1]

	imull	4(%ecx), %eax		C xp[1] * yp[1]

	imull	(%ecx), %edx		C xp[2] * yp[0]

	addl	%eax, %ebp
	popl	%esi

	addl	%ebp, %edx
	popl	%ebp

	movl	%edx, 8(%edi)		C wp[2]
	popl	%edi

	popl	%ebx

	ret


C -----------------------------------------------------------------------------

C defframe(SAVE_EBX,   -4)
C defframe(SAVE_ESI,   -8)
C defframe(SAVE_EDI,   -12)
C defframe(SAVE_EBP,   -16)
define(VAR_COUNTER,`PARAM_WP')
define(VAR_JMP,    `PARAM_SIZE')

	ALIGN(16)
L(four_or_more):
deflit(`FRAME',0)

	C eax	xp
	C ebx
	C ecx
	C edx

C First multiply yp[0]*xp[0..size] and store at dst[0..size].
C
C This loop is 8.0 c/l, compared to mpn_mul_1 6.25 c/l.

	pushl	%edi	defframe_pushl(SAVE_EDI)
	movl	PARAM_YP, %edx

	pushl	%ebx	defframe_pushl(SAVE_EDI)
	movl	PARAM_WP, %edi

	pushl	%esi	defframe_pushl(SAVE_EDI)
	movl	PARAM_SIZE, %ecx

	pushl	%ebp	defframe_pushl(SAVE_EDI)
	movl	%eax, %ebx	C xp

	movl	(%edx), %ebp	C yp[0] multiplier
	subl	$2, %ecx

	movl	(%eax), %eax	C xp[0]

	mull	%ebp

	movl	%eax, (%edi)	C wp[0]
	movl	%edx, %esi	C initial carry
	

	ALIGN(16)
L(mul_1):
	C eax	scratch
	C ebx	xp, incrementing
	C ecx	counter
	C edx	scratch
	C esi	carry
	C edi	wp, incrementing
	C ebp	multiplier

	movl	4(%ebx), %eax
	addl	$4, %ebx

	mull	%ebp

	addl	%esi, %eax
	movl	$0, %esi

	adcl	%edx, %esi

	movl	%eax, 4(%edi)
	addl	$4, %edi

	loop	L(mul_1)


	imull	4(%ebx), %ebp	C xp[size]

	addl	%esi, %ebp

	movl	%ebp, 4(%edi)	C wp[size]



C Addmul yp[n]*src[0..size-1-n] at dst[n...size-1], for each n=1..size-2.
C
C The last two addmuls, which are the bottom right corner of the product
C triangle, are left to the end.  These are src[size-3]*src[size-2,size-1]
C and src[size-2]*src[size-1].  If size is 4 then it's only these corner
C cases that need to be done.
C
C The unrolled code is the same as mpn_addmul_1(), see that routine for some
C comments.
C
C VAR_COUNTER is the outer loop, running from -(size-4) to -1, inclusive.
C
C VAR_JMP is the computed jump into the unrolled code, stepped by one code
C chunk each outer loop.
C
C K6 doesn't do any branch prediction on indirect jumps, which is good
C actually because it's a different target each time.  The unrolled addmul
C is about 3 cycles/limb faster than a simple loop, so the 6 cycle cost of
C the indirect jump is quickly recovered.


dnl  This value is also implicitly encoded in a shift and add.
dnl
deflit(CODE_BYTES_PER_LIMB, 15)

dnl  With the unmodified &src[size] and &dst[size] pointers, the
dnl  displacements in the unrolled code fit in a byte for UNROLL_COUNT
dnl  values up to 31.  Above that an offset must be added to them.
dnl
deflit(OFFSET,
ifelse(eval(UNROLL_COUNT>31),1,
eval((UNROLL_COUNT-31)*4),
0))

	C eax
	C ebx	xp+size-2
	C ecx
	C edx
	C esi
	C edi	wp+size-2
	C ebp

	movl	PARAM_SIZE, %ecx

	movl	PARAM_YP, %ebp
	leal	(%ebp,%ecx,4), %ebp	C &yp[size]

	subl	$3, %ecx		C size-3
	ASSERT(nc)

	movl	%ecx, %edx
ifelse(OFFSET,0,,
`	subl	$OFFSET, %ebx')

	shll	$4, %ecx
ifelse(OFFSET,0,,
`	subl	$OFFSET, %edi')

	negl	%ecx
	movl	%ebp, PARAM_YP		C &yp[size]

ifdef(`PIC',`
	call	L(pic_calc)
L(here):
',`
	leal	L(unroll_inner_end)-eval(CODE_BYTES_PER_LIMB)(%ecx,%edx), %ecx
')
	negl	%edx


	C The calculated jump mustn't be before the start of the available
	C code.  This is the limitation UNROLL_COUNT puts on the src operand
	C size, but checked here using the jump address directly.
	C
	ASSERT(ae,`
	movl_text_address( L(unroll_inner_start), %eax)
	cmpl	%eax, %ecx
	')


C -----------------------------------------------------------------------------
	ALIGN(16)
L(unroll_outer_top):
	C eax
	C ebx	&xp[size-2]
	C ecx	VAR_JMP
	C edx	VAR_COUNTER, limbs, negative
	C esi
	C edi	&wp[size-2]
	C ebp	&yp[size-2]

	movl	-8(%ebp,%edx,4), %ebp		C multiplier yp[i]
	movl	%edx, VAR_COUNTER

	movl	-4+OFFSET(%ebx,%edx,4), %eax	C xp[0]

	mull	%ebp

	testb	$1, %cl				C register order below

	movl	%edx, %esi			C high carry
	leal	CODE_BYTES_PER_LIMB(%ecx), %edx	C jump

	movl	%eax, %ecx			C low carry
	movl	%edx, VAR_JMP

	C A branch-free version of this using some xors was found to be a
	C touch slower than just a conditional jump, despite the jump
	C switching between taken and not taken on every loop.

ifelse(eval(UNROLL_COUNT%2),0,
	jz,jnz)	L(unroll_noswap)
	movl	%esi, %eax	C high,low carry other way around

	movl	%ecx, %esi
	movl	%eax, %ecx
L(unroll_noswap):

	jmp	*%edx


	C Must be on an even address here so the low bit of the jump address
	C will indicate which way around ecx/esi should start.
	C
	C An attempt was made at padding here to get the end of the unrolled
	C code to come out on a good alignment, to save padding before
	C L(corner).  This worked, but turned out to run slower than just an
	C ALIGN(2).  The reason for this is not clear, it might be related
	C to the different speeds on different UNROLL_COUNTs noted above.

	ALIGN(2)

L(unroll_inner_start):
	C eax	scratch
	C ebx	&xp[size-2-i]
	C ecx	carry low
	C edx	scratch
	C esi	carry high
	C edi	&wp[size-2]
	C ebp	multiplier
	C
	C 15 code bytes each limb
	C ecx/esi swapped on each chunk

forloop(`i', UNROLL_COUNT, 1, `
	deflit(`disp_src', eval(-i*4 + OFFSET))
	deflit(`disp_dst', eval(disp_src))

	m4_assert(`disp_src>=-128 && disp_src<128')
	m4_assert(`disp_dst>=-128 && disp_dst<128')

ifelse(eval(i%2),0,`
Zdisp(	movl,	disp_src,(%ebx), %eax)
	mull	%ebp
Zdisp(	addl,	%esi, disp_dst,(%edi))
	adcl	%eax, %ecx
	movl	%edx, %esi
	jadcl0( %esi)
',`
	dnl  this one comes out last
Zdisp(	movl,	disp_src,(%ebx), %eax)
	mull	%ebp
Zdisp(	addl,	%ecx, disp_dst,(%edi))
	adcl	%eax, %esi
	movl	%edx, %ecx
	jadcl0( %ecx)
')
')
L(unroll_inner_end):

	imull	m4_empty_if_zero(OFFSET)(%ebx), %ebp	C xp[size-i]*yp[i]

	addl	%esi, m4_empty_if_zero(OFFSET)(%edi)	C wp[size-2]

	adcl	%ebp, %ecx

	addl	%ecx, 4+OFFSET(%edi)	C wp[size-1]

	movl	VAR_COUNTER, %edx
	subl	$4, %ebx

	movl	VAR_JMP, %ecx
	incl	%edx

	movl	PARAM_YP, %ebp		C yp+size
	jnz	L(unroll_outer_top)


	C Final corner xp[0..1]*yp[size-1] and xp[0]*yp[size]
	C
	C eax
	C ebx	&xp[1]
	C ecx
	C edx
	C esi
	C edi	&wp[size-2]
	C ebp	&yp[size]

	movl	-4+OFFSET(%ebx), %eax			C xp[0]
	movl	%eax, %ecx

	movl	-8(%ebp), %esi				C yp[size-2]

	mull	%esi					C xp[0] * yp[size-2]

	imull	m4_empty_if_zero(OFFSET)(%ebx), %esi	C xp[1] * yp[size-2]

	imull	-4(%ebp), %ecx				C xp[0] * yp[size-1]

	addl	%eax, m4_empty_if_zero(OFFSET)(%edi)	C wp[size-2]

	adcl	%esi, %edx

	addl	%ecx, %edx
	popl	%ebp

	addl	%edx, 4+OFFSET(%edi)			C wp[size-1]

	popl	%esi
	popl	%ebx

	popl	%edi

	ret




	

ifelse(OFFSET,0,,`
	addl	$OFFSET, %ebx
	addl	$OFFSET, %edi
')


C -----------------------------------------------------------------------------
	ALIGN(16)
L(corner):
	C ebx	&src[size]
	C edi	&dst[2*size-5]

	movl	-12(%ebx), %ebp

	movl	-8(%ebx), %eax
	movl	%eax, %ecx

	mull	%ebp

	addl	%eax, -4(%edi)
	adcl	$0, %edx

	movl	-4(%ebx), %eax
	movl	%edx, %esi
	movl	%eax, %ebx

	mull	%ebp

	addl	%esi, %eax
	adcl	$0, %edx

	addl	%eax, (%edi)
	adcl	$0, %edx

	movl	%edx, %esi
	movl	%ebx, %eax

	mull	%ecx

	addl	%esi, %eax
	movl	%eax, 4(%edi)

	adcl	$0, %edx

	movl	%edx, 8(%edi)
	
	ret



C -----------------------------------------------------------------------------
ifdef(`PIC',`
L(pic_calc):
        C See mpn/x86/README about old gas bugs
	addl	(%esp), %ecx
	addl	$L(unroll_inner_end)-L(here)-eval(2*CODE_BYTES_PER_LIMB), %ecx
	addl	%edx, %ecx
	ret
')


EPILOGUE()
